#include "all.h"

#define DEBUG_LEVEL 0
//#define DEBUG_LEVEL 2

/*

backup/restore are to fork/exec to start this process.  This process
does not background itself because it doesn't need to, and it stays
in the parent's process group so that it will get killed should the
parent abort.

SMB vs CIFS?  Sigh.  Let's call it CIFS everywhere.

*/

#define REPORT_ERROR(err,a...) \
	{ \
		fprintf (fp, "ERROR " err "\n", ## a); \
		return; \
	}

#define REPORT_ERRNO       REPORT_ERROR ("%d", errno)

#define PARSE_FD(str) \
	{ \
		fd = (int) strtol (str, NULL, 10); \
		if (fd < SMBC_BASE_FD) \
			REPORT_ERROR ("Bad file descriptor"); \
	}

#define PARSE_LEN(str) \
	len = (uint64_t) strtoull (str, NULL, 10)

static void handle_open (const char *path, FILE *fp, int oflag)
{
	int fd;

	if (!*path)
		REPORT_ERROR ("no file");
	if ((fd = smbc_open (path, oflag, 0755)) < 0)
		REPORT_ERRNO;
	fprintf (fp, "OK %d\n", fd);
}

static void handle_close (const char *str, FILE *fp)
{
	int fd;

	PARSE_FD (str);
	if (smbc_close (fd))
		REPORT_ERRNO;
	fprintf (fp, "OK\n");
}

static void handle_read (const char *params, FILE *fp)
{
	const char *sep = " ";
	char *copy, *ctx, *fdstr, *lenstr, *buf;
	int fd;
	off_t len, ret;

	buf = NULL;
	copy = xstrdup (params);
	if (!(fdstr = strtok_r (copy, sep, &ctx)))
		goto error;
	if (!(lenstr = strtok_r (NULL, sep, &ctx)))
		goto error;
	PARSE_FD (fdstr);
	PARSE_LEN (lenstr);

	buf = xmalloc (len);
	if ((ret = smbc_read (fd, buf, len)) < 0)
		goto error;
	fprintf (fp, "OK %" PRIu64 "\n", ret);
	if (writeall (fileno (fp), buf, ret) < 0)
		pthread_exit (NULL);
	goto done;

	error:
	fprintf (fp, "ERROR arguments\n");

	done:
	free (copy);
	if (buf)
		free (buf);
}

static void handle_write (const char *params, FILE *fp)
{
	const char *sep = " ";
	char *copy, *ctx, *fdstr, *lenstr, *buf;
	int fd;
	off_t len, ret;

	buf = NULL;
	copy = xstrdup (params);
	if (!(fdstr = strtok_r (copy, sep, &ctx)))
		goto error;
	if (!(lenstr = strtok_r (NULL, sep, &ctx)))
		goto error;
	PARSE_FD (fdstr);
	PARSE_LEN (lenstr);

	buf = xmalloc (len);
	if (readall (fileno (fp), buf, len) < 0)
		pthread_exit (NULL);
	if ((ret = smbc_write (fd, buf, len)) < 0)
		goto error;
	fprintf (fp, "OK %" PRIu64 "\n", ret);
	goto done;

	error:
	fprintf (fp, "ERROR arguments\n");

	done:
	free (copy);
	if (buf)
		free (buf);
}

static void handle_lseek (const char *params, FILE *fp)
{
	const char *sep = " ";
	char *copy, *ctx, *fdstr, *offstr, *whencestr;
	int fd, whence;
	off_t offset, ret;

	copy = xstrdup (params);
	if (!(fdstr = strtok_r (copy, sep, &ctx)))
		goto error;
	if (!(offstr = strtok_r (NULL, sep, &ctx)))
		goto error;
	if (!(whencestr = strtok_r (NULL, sep, &ctx)))
		goto error;
	PARSE_FD (fdstr);
	offset = strtoull (offstr, NULL, 10);
	if (strlen (whencestr) != 1)
		goto error;
	switch (*whencestr)
	{
		case 's':
			whence = SEEK_SET;
			break;
		case 'c':
			whence = SEEK_CUR;
			break;
		case 'e':
			whence = SEEK_END;
			break;
		default:
			goto error;
	}
	if ((ret = smbc_lseek (fd, offset, whence)) < 0)
		goto error;
	fprintf (fp, "OK %" PRIu64 "\n", ret);
	goto done;

	error:
	fprintf (fp, "ERROR arguments\n");

	done:
	free (copy);
}

static void handle_size (const char *path, FILE *fp)
{
	struct stat buf;

	if (!*path)
		REPORT_ERROR ("no file");
	if (smbc_stat (path, &buf))
		REPORT_ERRNO;
	fprintf (fp, "OK %" PRId64 "\n", buf.st_size);
}

static void handle_unlink (const char *path, FILE *fp)
{
	if (!*path)
		REPORT_ERROR ("no file");
	if (smbc_unlink (path))
		REPORT_ERRNO;
	fprintf (fp, "OK\n");
}

static void handle_rename (const char *params, FILE *fp)
{
	char *a, *b, *end;
	int buflen, alen, blen, res;

	if (!*params)
		REPORT_ERROR ("no file");

	// these are more than we need
	buflen = strlen (params);
	a = xmalloc (buflen);
	b = xmalloc (buflen);

	// extract filename lengths and filenames
	alen = strtoul (params, &end, 10);
	if ((params == end) || (*end != ' '))
		REPORT_ERROR ("bad params");
	params = end + 1;
	blen = strtoul (params, &end, 10);
	if ((params == end) || (*end != ' '))
		REPORT_ERROR ("bad params");
	params = end + 1;
	if (strlen (params) != (alen + blen + 1))
		REPORT_ERROR ("bad params");
	memset (a, 0, buflen);
	memset (b, 0, buflen);
	strncpy (a, params, alen);
	strncpy (b, params + alen + 1, blen);

	// do the rename
	res = smbc_rename (a, b);
	free (a);
	free (b);

	if (res)
		REPORT_ERRNO;
	fprintf (fp, "OK\n");
}

static void handle_opendir (const char *path, FILE *fp)
{
	int fd;

	if (!*path)
		REPORT_ERROR ("no directory");
	if ((fd = smbc_opendir (path)) < 0)
		REPORT_ERRNO;
	fprintf (fp, "OK %d\n", fd);
}

static void handle_readdir (const char *params, FILE *fp)
{
	int fd;
	struct smbc_dirent *entry;

	if (!*params)
		REPORT_ERROR ("no directory");
	PARSE_FD (params);
	if (!(entry = smbc_readdir (fd)))
		fprintf (fp, "OK\n"); // done
	else
		fprintf (fp, "OK %s\n", entry->name);
}

static void handle_closedir (const char *params, FILE *fp)
{
	int fd;

	if (!*params)
		REPORT_ERROR ("no directory");
	PARSE_FD (params);
	smbc_closedir (fd);
	fprintf (fp, "OK\n");
}

static void handle_mkdir (const char *path, FILE *fp)
{
	if (!*path)
		REPORT_ERROR ("no directory");
	if (smbc_mkdir (path, 0700))
		REPORT_ERRNO;
	fprintf (fp, "OK\n");
}

static void handle_rmdir (const char *path, FILE *fp)
{
	if (!*path)
		REPORT_ERROR ("no directory");
	if (smbc_rmdir (path))
		REPORT_ERRNO;
	fprintf (fp, "OK\n");
}

static void handle_access (const char *path, FILE *fp)
{
	struct stat buf;

	if (!*path)
		REPORT_ERROR ("no path");
	if (smbc_stat (path, &buf))
		REPORT_ERRNO;
	fprintf (fp, "OK\n");
}

static void client_handler (int sock)
{
	char *buf, *params;
	int debug, len;
	FILE *fp;

	fp = fdopen (sock, "r+");
	setvbuf (fp, NULL, _IONBF, 0);
	debug = getenv ("CIFSD_DEBUG") != NULL;
	buf = xmalloc (len = 1024 * 1024);

	while (fgets (buf, len, fp))
	{
		if (debug)
			fprintf (stderr, ">>> %s", buf);

		// strip newline
		buf[strlen (buf) - 1] = 0;

		// quit is easy
		if (!strcasecmp (buf, CIFS_CMD_QUIT))
			return;

		// the rest have params
		params = strstr (buf, " ") + 1;

		if (!strncasecmp (buf, CIFS_CMD_OPEN_R " ", 7))
			handle_open (params, fp, O_RDONLY);
		else if (!strncasecmp (buf, CIFS_CMD_OPEN_W " ", 7))
			handle_open (params, fp, O_WRONLY | O_CREAT);
		else if (!strncasecmp (buf, CIFS_CMD_CLOSE " ", 6))
			handle_close (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_READ " ", 5))
			handle_read (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_WRITE " ", 6))
			handle_write (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_LSEEK " ", 6))
			handle_lseek (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_SIZE " ", 5))
			handle_size (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_UNLINK " ", 7))
			handle_unlink (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_RENAME " ", 7))
			handle_rename (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_OPENDIR " ", 8))
			handle_opendir (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_READDIR " ", 8))
			handle_readdir (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_CLOSEDIR " ", 9))
			handle_closedir (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_MKDIR " ", 6))
			handle_mkdir (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_RMDIR " ", 6))
			handle_rmdir (params, fp);
		else if (!strncasecmp (buf, CIFS_CMD_ACCESS " ", 7))
			handle_access (params, fp);
		else
			fprintf (fp, "ERROR\n");
	}
	fclose (fp);
}

static void get_auth_data (const char *srv, const char *shr, char *wg,
	int wglen, char *un, int unlen, char *pw, int pwlen)
{
	char *str;

	if ((str = getenv ("CIFS_DOMAIN")))
		strlcpy (wg, str, wglen);
	if ((str = getenv ("CIFS_USERNAME")))
		strlcpy (un, str, unlen);
	if ((str = getenv ("CIFS_PASSWORD")))
		strlcpy (pw, str, pwlen);
}

int main (void)
{
	SMBCCTX *smb_ctx;

	// setup smb client
	if (smbc_init (get_auth_data, DEBUG_LEVEL))
		ERROR ("smbc_init");
	smb_ctx = smbc_set_context (NULL);
	smbc_setOptionSmbEncryptionLevel (smb_ctx, SMBC_ENCRYPTLEVEL_REQUEST);

	// handle parent process
	client_handler (STDIN_FILENO);
	return 0;
}

